// BPM function based on DemoReel100.ino by Mark Kriegsman, December 2014 https://github.com/FastLED/FastLED/blob/master/examples/DemoReel100/DemoReel100.ino
#include <FastLED.h>
#include "MAX30105.h"
#include "heartRate.h"
MAX30105 particleSensor;

#define DATA_PIN   10
#define LED_TYPE    WS2811
#define COLOR_ORDER GRB
#define NUM_LEDS    150
CRGB leds[NUM_LEDS];

#define BRIGHTNESS          50
#define FRAMES_PER_SECOND  120

const byte RATE_SIZE = 4; //Increase this for more averaging of IR values. 4 is good.
byte rates[RATE_SIZE]; //Array of heart rates
byte rateSpot = 0;
long lastBeat = 0; //Time at which the last beat occurred

float beatsPerMinute;
int beatAvg;
const int NUM_READINGS = 5; //Adjust the number of readings before displaying Heart Rate
int valid_HR[NUM_READINGS];
int index_valid = 0;

int lastnonValidMillis = 0; // Variable to store the last non valid time
int noFinger_interval = 0; // Current interval between last not valid reading
int lastUpdateMillis = 0; // Variable to store the last update time
int current_interval = 0; // Current interval between last valid reading
int UPDATE_INTERVAL = 1000; // Update interval in milliseconds (adjust as needed)

uint8_t gHue = 0; // rotating "base color" used by many of the patterns
unsigned long lastSensorReadMillis = 0;
uint8_t BeatsPerMinute = 45;
uint8_t avgHeartRate = 0;
bool fingerContact = false;

void setup() {
  Serial.begin(115200);

  // Initialize sensor
  if (!particleSensor.begin(Wire, I2C_SPEED_FAST)) //Use default I2C port, 400kHz speed
  {
    Serial.println("MAX30105 was not found. Please check wiring/power. ");
    while (1);
  }

  particleSensor.setup(); //Configure sensor with default settings
  particleSensor.setPulseAmplitudeRed(0x0A); //Turn Red LED to low to indicate sensor is running

  // Initialize valid HR reading array
  for (int i = 0; i < NUM_READINGS; i++) {
    valid_HR[i] = 0;
  }

  delay(3000); // 3 second delay for recovery

  // tell FastLED about the LED strip configuration
  FastLED.addLeds<LED_TYPE, DATA_PIN, COLOR_ORDER>(leds, NUM_LEDS).setCorrection(TypicalLEDStrip);
  fill_solid(leds, NUM_LEDS, CRGB(0, 0, 0));       // fill all black
  FastLED.show();                                   // show
  // set master brightness control
  FastLED.setBrightness(BRIGHTNESS);
}


void loop()
{
  avgHeartRate = getHeartRate();
  if (millis() - lastSensorReadMillis >= 5000) {
    BeatsPerMinute = avgHeartRate; 
   Serial.print(" LED input is ");   
   Serial.println(BeatsPerMinute);   
    lastSensorReadMillis = millis(); // Update the last sensor read time
  }

  if (fingerContact){
  // Call the current pattern function once, updating the 'leds' array
  bpm(BeatsPerMinute);
  }
  else{
    sinelon();
  }
  
  // send the 'leds' array out to the actual LED strip
  FastLED.show();  
  // insert a delay to keep the framerate modest
  FastLED.delay(1000/FRAMES_PER_SECOND); 

  // do some periodic updates
  EVERY_N_MILLISECONDS( 50 ) { gHue++; } // slowly cycle the "base color" through the rainbow
}

void sinelon()
{
  // a colored dot sweeping back and forth, with fading trails
  fadeToBlackBy( leds, NUM_LEDS, 20);
  int pos = beatsin16( 13, 0, NUM_LEDS-1 );
  leds[pos] += CRGB(255, 0, 0);
}

void bpm (uint8_t BeatsPerMinute )
{
  // colored stripes pulsing at a defined Beats-Per-Minute (BPM)
  CRGBPalette16 palette3 = OceanColors_p;
  CRGBPalette16 palette2 = RainbowColors_p;
  CRGBPalette16 palette1 = LavaColors_p;

  uint8_t beat = beatsin8( BeatsPerMinute, 10, 255);
  for ( int i = 0; i < NUM_LEDS; i++) { //9948
    //leds[i] = ColorFromPalette(palette, gHue+(i*2), beat-gHue+(i*10));
    if (BeatsPerMinute <= 60) {
      leds[i] = ColorFromPalette(palette3, gHue + (i * 2), beat );
    }
    else if (BeatsPerMinute > 60 && BeatsPerMinute <= 90) {
      leds[i] = ColorFromPalette(palette2, gHue + (i * 2), beat );
       addGlitter(10);
    }
    else if (BeatsPerMinute > 90) {
      leds[i] = ColorFromPalette(palette1, gHue + (i * 2), beat );
      addGlitter(1);
    }
  }
  
    // send the 'leds' array out to the actual LED strip
  FastLED.show();
  // insert a delay to keep the framerate modest
  FastLED.delay(1000 / FRAMES_PER_SECOND);

  EVERY_N_MILLISECONDS( 50 ) {gHue++;  }// slowly cycle the "base color" through the rainbow
   /*// Define the beat based on the BPM
  uint8_t beat = beatsin8(BeatsPerMinute, 64, 255);

  // Calculate brightness based on the beat value
  uint8_t brightness = map(beat, 64, 255, 10, 255);

  // Fill all LEDs with the same color and adjust brightness
  fill_solid(leds, NUM_LEDS, CRGB(BeatsPerMinute*3, BeatsPerMinute*3, BeatsPerMinute*3)); // Adjust the RGB values as needed
  //fill_solid(leds, NUM_LEDS, CRGB(255, 255, 255)); // Adjust the RGB values as needed
  FastLED.setBrightness(brightness);*/
}

void addGlitter( fract8 chanceOfGlitter)
{
  if ( random8() < chanceOfGlitter) {
    leds[ random16(NUM_LEDS) ] += CRGB::White;
  }
}


uint8_t getHeartRate() {
  long irValue = particleSensor.getIR();

  if (checkForBeat(irValue) == true)
  {
    //We sensed a beat!
    long delta = millis() - lastBeat;
    lastBeat = millis();

    beatsPerMinute = 60 / (delta / 1000.0);

    if (beatsPerMinute < 220 && beatsPerMinute > 30)
    {
      rates[rateSpot++] = (byte)beatsPerMinute; //Store this reading in the array
      rateSpot %= RATE_SIZE; //Wrap variable

      //Take average of readings
      beatAvg = 0;
      for (byte x = 0 ; x < RATE_SIZE ; x++)
        beatAvg += rates[x];
      beatAvg /= RATE_SIZE;
    }
  }

  // Update the running average or non valid messages at regular intervals
  noFinger_interval = millis() - lastnonValidMillis;
  current_interval = millis() - lastUpdateMillis;
  if (irValue < 50000 && noFinger_interval >= 1000) {           // no valid reading
    lastnonValidMillis = millis();
    fingerContact = false;
    Serial.println(" No finger?");
    //fadeToBlackBy( leds, NUM_LEDS, 20);   
  }

  else if (irValue >= 50000 && current_interval >= UPDATE_INTERVAL && beatAvg < 125 && beatAvg > 0) { // sensing a presence, will determine if reading is valid
    fingerContact = true;
    lastUpdateMillis = millis();

    valid_HR[index_valid] = beatAvg;
    index_valid = (index_valid + 1) % NUM_READINGS;    // make sure you will only fill array until NUM_READINGS

    Serial.print("Valid BPMs are= ["); Serial.print(valid_HR[0] ); Serial.print("] [");
    Serial.print(valid_HR[1] ); Serial.print("] ["); Serial.print(valid_HR[2] ); Serial.print("] [");
    Serial.print(valid_HR[3] ); Serial.print("] ["); Serial.print(valid_HR[4] ); Serial.print("]");
    Serial.println();

    // Calculate the average and difference between max and min
    int sum = 0;

    for (int i = 0; i < NUM_READINGS; i++) {
      sum += valid_HR[i];

    }
    avgHeartRate = sum / NUM_READINGS;
    Serial.print ("Current average is ");
    Serial.println (avgHeartRate);
  }
    return avgHeartRate;
}
